from abc import ABC
from typing import Dict

from .fields import BaseField, GraphField, IntegerField


class EntityMeta(type):
    def __new__(cls, name, bases, dct):
        """
        Annotate new Entity class with 'fields' attributes.
        """
        new_cls = super().__new__(cls, name, bases, dct)
        fields = EntityMeta._get_fields(dct)

        for base_cls in bases:
            base_fields = EntityMeta._get_fields(base_cls.__dict__)
            for field_name in base_fields.keys():
                msg = f"Cannot define field '{field_name}' on class '{name}' because it is already fined in parent class {base_cls}"
                assert field_name not in fields, msg

            fields = {**fields, **base_fields}

        new_cls.fields = fields
        return new_cls

    @staticmethod
    def _get_fields(dct: dict):
        fields = {}
        for name, field in dct.items():
            if issubclass(field.__class__, BaseField):
                fields[name] = field

        return fields


class BaseEntity(metaclass=EntityMeta):
    fields: Dict[str, BaseField]

    def __init__(self, **kwargs):
        for field_name, field in self.fields.items():
            field_val = kwargs.get(field_name)
            if not field_val and field.default is not None:
                field_val = field.default
            elif not field_val and field.distribution is not None:
                field_val = field.distribution()
            elif not field_val:
                raise ValueError(f"Field {field_name} does not have a value.")

            assert field_val is not None
            field.validate(field_val)
            setattr(self, field_name, field_val)

    @classmethod
    def assert_fieldname(cls, field_name):
        assert (
            field_name in cls.fields.keys()
        ), f"Could not find field {field_name} in entity {cls}."


class BaseAgent(BaseEntity):
    pass


class BaseNetwork(BaseEntity):
    graph = GraphField()
    size = IntegerField(default=0)



"""
NetworX graph visualization layouts
"""
import math
from random import random, seed

import networkx as nx

P = 0.05


def f_a(d, k):
    """Returns the attractive force"""
    return d * d / k


def f_r(d, k):
    """Returns the repulsive force"""
    return 10 * k * k / d


def fruchterman_reingold_layout(
    G: nx.Graph,
    size: int,
    iteration: int = 50,
    layout_seed: int = 1,
):
    """
    Generates a NetworkX compatiable graph layout using the Fructerman + Reingold graph layout algorithm
    https://en.wikipedia.org/wiki/Force-directed_graph_drawing

    Originally posted by  mmisono at
    https://gist.github.com/mmisono/8972731
    """
    length, width = size, size
    seed(layout_seed)

    area = width * length
    k = math.sqrt(area / nx.number_of_nodes(G))

    # initial position
    for node in G.nodes.values():
        node["x"] = width * random()
        node["y"] = length * random()

    t = width / 10
    dt = t / (iteration + 1)

    for _ in range(iteration):
        pos = {}
        for v, node in G.nodes.items():
            pos[v] = [node["x"], node["y"]]

        # calculate repulsive forces
        for v, node in G.nodes.items():
            node["dx"] = 0
            node["dy"] = 0
            for u in G.nodes.keys():
                if v != u:
                    dy = node["y"] - G.nodes[u]["y"]
                    dx = node["x"] - G.nodes[u]["x"]
                    delta = math.sqrt(dx * dx + dy * dy)
                    if delta != 0:
                        d = f_r(delta, k) / delta
                        node["dx"] += dx * d
                        node["dy"] += dy * d

        # calculate attractive forces
        for v, u in G.edges.keys():
            dx = G.nodes[v]["x"] - G.nodes[u]["x"]
            dy = G.nodes[v]["y"] - G.nodes[u]["y"]
            delta = math.sqrt(dx * dx + dy * dy)
            if delta != 0:
                d = f_a(delta, k) / delta
                ddx = dx * d
                ddy = dy * d
                G.nodes[v]["dx"] += -ddx
                G.nodes[u]["dx"] += +ddx
                G.nodes[v]["dy"] += -ddy
                G.nodes[u]["dy"] += +ddy

        # limit the maximum displacement to the temperature t
        # and then prevent from being displace outside frame
        for v in G.nodes.keys():
            dx = G.nodes[v]["dx"]
            dy = G.nodes[v]["dy"]
            disp = math.sqrt(dx * dx + dy * dy)
            if disp != 0:
                d = min(disp, t) / disp
                x = G.nodes[v]["x"] + dx * d
                y = G.nodes[v]["y"] + dy * d
                x = min(width, max(0, x)) - width / 2
                y = min(length, max(0, y)) - length / 2
                G.nodes[v]["x"] = (
                    min(
                        math.sqrt(width * width / 4 - y * y),
                        max(-math.sqrt(width * width / 4 - y * y), x),
                    )
                    + width / 2
                )
                G.nodes[v]["y"] = (
                    min(
                        math.sqrt(length * length / 4 - x * x),
                        max(-math.sqrt(length * length / 4 - x * x), y),
                    )
                    + length / 2
                )

        # cooling
        t -= dt

    pos = {}
    for v in G.nodes.keys():
        pos[v] = [G.nodes[v]["x"], G.nodes[v]["y"]]

    return pos



import random
from typing import List, Optional

import numpy as np

from .entities import BaseEntity, BaseNetwork


class Registry:
    """
    Holds data about an entity type and is used to query that data.
    """

    def __init__(self, cls: BaseEntity, initial_number: int = 0):
        self.cls = cls
        self.vals = {}
        self._is_network = issubclass(cls, BaseNetwork)
        self._initial_number = initial_number
        self.reset()

    def reset(self):
        """
        Reset all fields
        """
        self.count = self._initial_number
        for field_name, field in self.cls.fields.items():
            self.vals[field_name] = field.setup(self._initial_number)

    @property
    def query(self):
        return Query(self)

    def add(self, entity: BaseEntity, id: Optional[int] = None) -> int:
        assert getattr(entity, "id", None) is None, "Cannot create an entity with an existing ID."
        entity_id = self.count
        entity.id = entity_id
        self.count += 1
        for field_name in self.cls.fields.keys():
            # Add each field to its value array.
            val = getattr(entity, field_name)
            is_obj_arr = self.vals[field_name].dtype == np.dtype("O")
            if is_obj_arr:
                # This is really dumb but you can't construct a NumPy object array with
                # only empty nx.Graph() objects - possibly because they're falsey,
                # so I put in this True value then immediately throw it away.
                # Please fix this if you can.
                els = self.vals[field_name].tolist() + [val, True]
                self.vals[field_name] = np.array(els, dtype=object)[:-1].copy()

            else:
                self.vals[field_name] = np.append(self.vals[field_name], [val])

        return entity_id

    def get(self, entity_id: int) -> BaseEntity:
        """
        Returns an instance of the entity's class, populated with the data form its compoenent.
        """
        assert entity_id < self.count, f"Entity ID {entity_id} out of range"
        entity_values = {k: self.vals[k][entity_id] for k in self.cls.fields.keys()}
        entity = self.cls(**entity_values)
        entity.id = entity_id
        return entity

    def write(self, entity_id: int, field_name: str, value) -> BaseEntity:
        """
        Sets a value on a particular entity.
        """
        assert entity_id < self.count, f"Entity ID {entity_id} out of range"
        self.vals[field_name][entity_id] = value

    def save(self, enitites: List[BaseEntity], field_names=None):
        """
        Save a list of entities to the registry.
        """
        fields_to_update = field_names if field_names is not None else self.cls.fields.keys()
        for entity in enitites:
            for field_name in fields_to_update:
                self.vals[field_name][entity.id] = getattr(entity, field_name)

    def add_node(self, entity_id: int, node_id: int):
        """
        Network only method.
        Add a node to the network's graph.
        Assumes the network is fully connected.
        """
        assert self._is_network, "This method can only be used for Network entities"
        assert entity_id < self.count, f"Entity ID {entity_id} out of range"
        # Add the node to the graph
        graph = self.vals["graph"][entity_id]
        assert not node_id in graph, f"Agent {node_id} already in {self.cls.__name__} {entity_id}"
        graph.add_node(node_id)
        edges = []
        for dest_id in graph.nodes:
            if dest_id != node_id:
                edges.append((node_id, dest_id))

        graph.add_edges_from(edges)

        # Update the size
        self.vals["size"][entity_id] += 1

    def remove_node(self, entity_id: int, node_id: int):
        """
        Network only method.
        Remove an node from the graph.
        Assumes the network is fully connected.
        """
        assert self._is_network, "This method can only be used for Network entities"
        assert entity_id < self.count, f"Entity ID {entity_id} out of range"
        graph = self.vals["graph"][entity_id]
        assert node_id in graph, f"Node {node_id} not in {self.cls} {entity_id}"
        graph.remove_node(node_id)

    def get_nodes(self, entity_id: int) -> np.ndarray:
        """
        Returns the node IDs of a  network.
        """
        assert self._is_network, "This method can only be used for Network entities"
        assert entity_id < self.count, f"Entity ID {entity_id} out of range"
        graph = self.vals["graph"][entity_id]
        return np.array(list(graph.nodes.keys()), dtype=np.int64)

    def get_node_contacts(self, entity_id: int, node_id: int):
        """
        Returns the direct contacts/network neighbours of a node.
        """
        assert self._is_network, "This method can only be used for Network entities"
        assert entity_id < self.count, f"Entity ID {entity_id} out of range"
        graph = self.vals["graph"][entity_id]
        return list(graph.neighbors(node_id))


class Query:
    FILTER_STRATEGIES = (
        # Greater than
        ("__gt", lambda a, b: a > b),
        # Less than
        ("__lt", lambda a, b: a < b),
        # Greater than or equal to
        ("__gte", lambda a, b: a >= b),
        # Less than or equal to
        ("__lte", lambda a, b: a <= b),
        # Not equal
        ("__ne", lambda a, b: a != b),
        # Equal
        ("", lambda a, b: a == b),
    )

    def __init__(self, registry: Registry, ids: Optional[np.array] = None):
        self._registry = registry
        if ids is None:
            self._ids = np.array(range(self._registry.count), dtype=np.int64)
        elif type(ids) is np.ndarray and ids.dtype is np.int64:
            self._ids = ids
        else:
            self._ids = np.array(ids, dtype=np.int64)

    def filter(self, **kwargs):
        """
        Returns a Query with entities filtered according to provided arguments.

        A filter kwarg ending with one of the following suffixes with use a different comparison strategy.
        Strategies:
            - __gt:  Greater than
            - __lt:  Less than
            - __gte: Greater than equal
            - __lte: Less than equal
            - __ne:  Not equal

        Example: "select all entities with an age greater than 12 and weight equal to 1"

        query.filter(age__gt=12, weight=1)

        """
        filtered_ids = self._filter(**kwargs)
        new_ids = np.intersect1d(self._ids, filtered_ids)
        return Query(self._registry, new_ids)

    def exclude(self, **kwargs):
        """
        Opposite of filter - throw away matching ids.
        """
        filtered_ids = self._filter(**kwargs)
        new_ids = np.setdiff1d(self._ids, filtered_ids)
        return Query(self._registry, new_ids)

    def _filter(self, **kwargs):
        mask = None
        for condition, value in kwargs.items():
            m = None
            for suffix, strategy in Query.FILTER_STRATEGIES:
                if condition.endswith(suffix):
                    c = condition[: (-1 * len(suffix))] if suffix else condition
                    self._registry.cls.assert_fieldname(c)
                    m = strategy(self._registry.vals[c], value)
                    break

            mask = m if mask is None else (mask * m)

        # Filter IDs based on mask
        return np.where(mask)[0]

    def choose(self, num_choices: int):
        """
        Returns a Query with entities filtered down to `num_choices` randomly chosen entities.
        """
        new_ids = random.choices(self._ids, k=num_choices)
        return Query(self._registry, new_ids)

    def select(self, entity_ids: List[int]):
        """
        Returns a Query with entities filtered down to those provided as `entity_ids`.
        """
        new_ids = np.intersect1d(self._ids, entity_ids)
        return Query(self._registry, new_ids)

    def deselect(self, entity_ids: List[int]):
        """
        Returns a Query with entities filtered down to those NOT provided as `entity_ids`.
        """
        new_ids = np.setdiff1d(self._ids, entity_ids)
        return Query(self._registry, new_ids)

    def where(self, func):
        """
        Returns the agent ids where the provided func evalutates to True.
        Function may be vectorized or request entity specific arguments.
        """
        if self._is_vectorized(func):
            # Get the mask with the vectorized function.
            args = self._get_requested_args(func)
            mask = func(*args)
        else:
            # Get the mask with the scalar function.
            mask = np.empty(len(self._ids), dtype=np.bool)
            for idx, entity_id in enumerate(self._ids):
                args = self._get_requested_args(func, entity_id=entity_id)
                mask[idx] = func(*args)

        new_ids = self._ids[mask]
        return Query(self._registry, new_ids)

    def update(self, **kwargs):
        """
        Update all selected entries with provided data.
        Function may be vectorized or request entity specific arguments.
        """
        for field_name, value in kwargs.items():
            self._registry.cls.assert_fieldname(field_name)
            if callable(value) and self._is_vectorized(value):
                # Update the fields according to the vectorized function.
                args = self._get_requested_args(value)
                update_values = value(*args)
            elif callable(value):
                # Update the fields according to the scalar function.
                update_values = np.empty(
                    len(self._ids), dtype=self._registry.vals[field_name].dtype
                )
                for idx, entity_id in enumerate(self._ids):
                    args = self._get_requested_args(value, entity_id=entity_id)
                    update_values[idx] = value(*args)
            else:
                # Just set everything to the single, static value.
                update_values = value

            self._registry.vals[field_name][self._ids] = update_values

    def all(self) -> List[BaseEntity]:
        """
        Returns all entities selected by the query as class instances.
        """
        entities = []
        for entity_id in self._ids:
            entities.append(self._registry.get(entity_id))

        return entities

    def only(self) -> BaseEntity:
        """
        Returns the class instance of the only remaining selected entity.
        """
        assert len(self._ids) == 1, "Can only use `only()` when one id is selected."
        return self._registry.get(self._ids[0])

    def id(self) -> List[int]:
        """
        Returns the ID of the only remaining selected entity.
        """
        assert len(self._ids) == 1, "Can only use `id()` when one id is selected."
        return self._ids[0]

    def ids(self) -> List[int]:
        return list(self._ids)

    def _get_requested_args(self, func, entity_id: Optional[int] = None):
        """
        Returns the arguments requested using the `@arguments` decorator.
        """
        arg_names = getattr(func, "_arg_names", [])
        return [self._get_requested_arg(n, entity_id) for n in arg_names]

    def _get_requested_arg(self, field_name, entity_id: Optional[int]):
        if entity_id is None:
            # Return vectorized argument for all selected ids.
            if field_name == "id":
                return self._ids
            else:
                return self._registry.vals[field_name][self._ids]
        else:
            # Return scalar argument for the given id.
            if field_name == "id":
                return entity_id
            else:
                return self._registry.vals[field_name][entity_id]

    def _is_vectorized(self, func):
        """
        Returns True if the given function is marked using @vectorized,
        which means it can accept array arguments and returns arrays.
        """
        return bool(getattr(func, "_vectorized", None))


def vectorized(func):
    """
    Decorator which marks an update function as vectorized.
    """
    func._vectorized = True
    return func


def arguments(*arg_names):
    """
    Decorator which defines the arguments for an update function.
    """

    def decorator(func):
        func._arg_names = arg_names
        return func

    return decorator



import numpy as np
import random

from .entities import BaseAgent, BaseNetwork
from .registry import Registry


class AgentModel:
    agents: Registry
    networks: Registry

    def __init__(
        self,
        start_time: float,
        end_time: float,
        timestep: float,
    ):
        # Setup run timesteps
        assert end_time > start_time, "End time must be greater than start time"
        time_period = end_time - start_time
        num_steps = 1 + (time_period / timestep)
        msg = f"Time step {timestep} must be less than time period {time_period}"
        assert num_steps >= 1, msg
        msg = f"Time step {timestep} must be a factor of time period {time_period}"
        assert num_steps % 1 == 0, msg
        self._times = np.linspace(start_time, end_time, num=int(num_steps))
        self._timestep = timestep

        self._setup_systems = []
        self._runtime_systems = []

        self.agents = None
        self.networks = None

        self._post_setup_hook = None
        self._post_timestep_hook = None

    def add_setup_step(self, func):
        self._setup_systems.append(func)

    def add_runtime_step(self, func):
        self._runtime_systems.append(func)

    def set_agent_class(self, agent_class: BaseAgent, initial_number: int = 0):
        self.agents = Registry(agent_class, initial_number)

    def set_network_class(self, network_class: BaseNetwork, initial_number: int = 0):
        self.networks = Registry(network_class, initial_number)

    def add_setup_hook(self, hook):
        """
        Register a function which will be run on the model just after setup
        """
        self._post_setup_hook = hook

    def add_timestep_hook(self, hook):
        """
        Register a function which will be run on the model after each timestep
        """
        self._post_timestep_hook = hook

    def run(self):
        assert self.agents, "An agent class must be set before running the model."
        assert self.networks, "A network class must be set before running the model."
        self._reset()
        self._run_setup()
        if self._post_setup_hook:
            self._post_setup_hook(self)

        for time in self._times:
            self._run_timestep(time)
            if self._post_timestep_hook:
                self._post_timestep_hook(self, time)

    def _reset(self):
        self.agents.reset()
        self.networks.reset()

    def _run_setup(self):
        for setup_step in self._setup_systems:
            setup_step(self)

    def _run_timestep(self, time):
        for runtime_step in self._runtime_systems:
            runtime_step(self, time)



from typing import Tuple

import numpy as np
import numba

import summer.flows as flows
from summer.compute import (
    accumulate_positive_flow_contributions,
    accumulate_negative_flow_contributions,
    sparse_pairs_accum,
    find_sum
)

from .model_runner import ModelRunner


class VectorizedRunner(ModelRunner):
    """
    An optimized, but less accessible model runner.
    """

    def __init__(self, model):
        super().__init__(model)

    def prepare_to_run(self):
        """Do all precomputation here"""

        # Disable caching; we don't want it, but the reference runner probably still does...
        self.model._enable_cache = False

        super().prepare_to_run()
        
        self.infectious_flow_indices = np.array([
            i for i, f in self._iter_non_function_flows if isinstance(f, flows.BaseInfectionFlow)
        ], dtype=int)
        self.death_flow_indices = np.array([
            i for i, f in self._iter_non_function_flows if f.is_death_flow
        ], dtype=int)
        
        # Include dummy values in population_idx to account for Entry flows
        non_func_pops = np.array(
            [f.source.idx if f.source else 0 for i, f in self._iter_non_function_flows], dtype=int
        )

        func_pops = np.array(
            [f.source.idx if f.source else 0 for i, f in self._iter_function_flows], dtype=int
        )

        self.population_idx = np.concatenate((non_func_pops, func_pops))

        # Store indices of flows that are not population dependent
        self._non_pop_flow_idx = np.array([i for i, f in self._iter_non_function_flows \
            if (type(f) in (flows.ReplacementBirthFlow, flows.ImportFlow))], dtype=int)
        self._has_non_pop_flows = bool(len(self._non_pop_flow_idx))

        # Crude birth flows use population sum rather than a compartment; store indices here
        self._crude_birth_idx = np.array([i for i, f in self._iter_non_function_flows \
            if type(f) == flows.CrudeBirthFlow], dtype=int)
        self._has_crude_birth = bool(len(self._crude_birth_idx))

        self._has_replacement = False
        # Replacement flows must be calculated after death flows, store indices here
        for i, f in self._iter_non_function_flows:
            if type(f) == flows.ReplacementBirthFlow:
                self._has_replacement = True
                self._replacement_flow_idx = i

        self._precompute_flow_weights()
        self._precompute_flow_maps()

        self._build_infectious_multipliers_lookup()

        # Initialize arrays for infectious multipliers
        # FIXME Put this in its own function, this is getting messy again
        num_cats = self.num_categories
        num_strains = len(self.model._disease_strains)
        self._infection_density = {}
        self._infection_frequency = {}
        self._infection_density_flat = np.empty((num_strains, num_cats), dtype=float)
        self._infection_frequency_flat = np.empty((num_strains, num_cats), dtype=float)

    def _precompute_flow_weights(self):
        """Calculate all static flow weights before running, and build indices for time-varying weights"""
        self.flow_weights = np.zeros(len(self.model._flows))
        time_varying_weight_indices = []
        for i, f in self._iter_non_function_flows:
            weight_type = f.weight_type()
            if weight_type == flows.WeightType.STATIC:
                weight = f.get_weight_value(0,None)
                self.flow_weights[i] = weight
            elif weight_type == flows.WeightType.FUNCTION:
                time_varying_weight_indices.append(i)
            # else: Is system, these are calculated in a separate process
            # FIXME Refactor so all in one place, maybe?

        self.time_varying_weight_indices = np.array(time_varying_weight_indices, dtype=int)

        self._map_blocks()

    def _map_blocks(self):
        flow_block_maps = {}
        for i in self.time_varying_weight_indices:
            f = self.model._flows[i]
            key = (f.param, tuple(f.adjustments))
            if key not in flow_block_maps:
                flow_block_maps[key] = []
            flow_block_maps[key].append(i)
        self.flow_block_maps = dict([(k, np.array(v,dtype=int)) for k,v in flow_block_maps.items()])


    def _precompute_flow_maps(self):
        """Build fast-access arrays of flow indices"""
        f_pos_map = []
        f_neg_map = []
        for i, f in self._iter_non_function_flows:
            if f.source:
                f_neg_map.append((i, f.source.idx))
            if f.dest:
                f_pos_map.append((i, f.dest.idx))
        for i, f in self._iter_function_flows:
            if f.source:
                f_neg_map.append((i, f.source.idx))
            if f.dest:
                f_pos_map.append((i, f.dest.idx))

        self._pos_flow_map = np.array(f_pos_map, dtype=np.int)
        self._neg_flow_map = np.array(f_neg_map, dtype=np.int)

    def _prepare_time_step(self, time: float, compartment_values: np.ndarray):
        """
        Pre-timestep setup. This should be run before `_get_rates`.
        Here we set up any stateful updates that need to happen before we get the flow rates.
        """
        
        # Some flows (e.g birth replacement) expect this value to be defined 
        self._timestep_deaths = 0.

        # Find the effective infectious population for the force of infection (FoI) calculations.
        mixing_matrix = self._get_mixing_matrix(time)

        self._calculate_strain_infection_values(compartment_values, mixing_matrix)

    def _apply_flow_weights_at_time(self, time, computed_values):
        """Calculate time dependent flow weights and insert them into our weights array

        Args:
            time (float): Time in model.times coordinates
        """

        # Run all adjustment systems; these are vectorized flow weight systems that operate over an array of flows
        for (s,flow_idx) in self._adjustment_system_flow_maps:
            self.flow_weights[flow_idx] = s.get_weights_at_time(time, computed_values)

        # Compute flow value blocks; these are 'chunked' blocks of flows that have a common flow param/adjustment structure,
        # but are still specified as operating on individual flows (as opposed to AdjustmentSystems, which are vectorized)
        for (param, adjustments), flow_idx in self.flow_block_maps.items():
            value = param(time, computed_values) if callable(param) else param
            for a in adjustments:
                value = a.get_new_value(value, computed_values, time)
            self.flow_weights[flow_idx] = value

    def _get_infectious_multipliers(self) -> np.ndarray:
        """Get multipliers for all infectious flows

        Returns:
            np.ndarray: Array of infectiousness multipliers
        """
        # We can use a fast lookup version if we have only one type of infectious multiplier 
        # (eg only frequency, not mixed freq and density)
        if self._infection_frequency_only:
            return self._infection_frequency_flat[self._infect_strain_lookup_idx, self._infect_cat_lookup_idx]
        elif self._infection_density_only:
            return self._infection_density_flat[self._infect_strain_lookup_idx, self._infect_cat_lookup_idx]

        multipliers = np.empty(len(self.infectious_flow_indices))
        for i, idx in enumerate(self.infectious_flow_indices):
            f = self.model._flows[idx]
            multipliers[i] = f.find_infectious_multiplier(f.source, f.dest)
        return multipliers

    def _build_infectious_multipliers_lookup(self):
        """Get multipliers for all infectious flows

        These are used by _get_infectious_multipliers_flat (currently experimental)

        Returns:
            np.ndarray: Array of infectiousness multipliers
        """
        lookups = []

        has_freq = False
        has_dens = False

        for i, idx in enumerate(self.infectious_flow_indices):
            f = self.model._flows[idx]
            if isinstance(f, flows.InfectionFrequencyFlow):
                has_freq = True
            elif isinstance(f, flows.InfectionDensityFlow):
                has_dens = True
            cat_idx, strain = self._get_infection_multiplier_indices(f.source, f.dest)
            strain_idx = self.model._disease_strains.index(strain)
            lookups.append([strain_idx, cat_idx])
        full_table = np.array(lookups, dtype=int)
        self._full_table = full_table.reshape(len(self.infectious_flow_indices),2)
        self._infect_strain_lookup_idx = self._full_table[:,0].flatten()
        self._infect_cat_lookup_idx = self._full_table[:,1].flatten()

        self._infection_frequency_only = False
        self._infection_density_only = False

        if has_freq and not has_dens:
            self._infection_frequency_only = True
        elif has_dens and not has_freq:
            self._infection_density_only = True

    def _calculate_strain_infection_values(
        self, compartment_values: np.ndarray, mixing_matrix: np.ndarray
    ):
        num_cats = self.num_categories
        # Calculate total number of people per category (for FoI).
        # A vector with size (num_cats).
        self._category_populations = sparse_pairs_accum(
            self._compartment_category_map, compartment_values, num_cats
        )

        for strain_idx, strain in enumerate(self.model._disease_strains):
            strain_compartment_infectiousness = self._compartment_infectiousness[strain]

            infection_density, infection_frequency = get_strain_infection_values(compartment_values,
                strain_compartment_infectiousness, self._compartment_category_map, num_cats, mixing_matrix,
                self._category_populations)
            
            self._infection_density[strain] = infection_density
            self._infection_density_flat[strain_idx] = infection_density

            self._infection_frequency[strain] = infection_frequency
            self._infection_frequency_flat[strain_idx] = infection_frequency

    def _get_flow_rates(self, compartment_vals: np.ndarray, time: float) -> np.ndarray:
        """Get current flow rates, equivalent to calling get_net_flow on all (non-function) flows

        Args:
            comp_vals (np.ndarray): Compartment values
            time (float): Time in model.times coordinates

        Returns:
            np.ndarray: Array of all (non-function) flow rates
        """

        computed_values = self._calc_computed_values(compartment_vals, time)

        self._apply_flow_weights_at_time(time, computed_values)

        # These will be filled in afterwards
        populations = compartment_vals[self.population_idx]
        # Update for special cases (population-independent and CrudeBirth)
        if self._has_non_pop_flows:
            populations[self._non_pop_flow_idx] = 1.0
        if self._has_crude_birth:
            populations[self._crude_birth_idx] = find_sum(compartment_vals)

        flow_rates = self.flow_weights * populations
        
        # Calculate infection flows
        infect_mul = self._get_infectious_multipliers()
        flow_rates[self.infectious_flow_indices] *= infect_mul

        # ReplacementBirthFlow depends on death flows already being calculated; update here
        if self._has_replacement:
            # Only calculate timestep_deaths if we use replacement, it's expensive...
            self._timestep_deaths = flow_rates[self.death_flow_indices].sum()
            flow_rates[self._replacement_flow_idx] = self._timestep_deaths


        self._apply_function_flow_rates(compartment_vals, flow_rates, computed_values, time)

        return flow_rates

    def _apply_function_flow_rates(self, comp_vals, flow_rates, computed_values, time):
        if self._iter_function_flows:
            # Evaluate the function flows.
            for flow_idx, flow in self._iter_function_flows:
                net_flow = flow.get_net_flow(
                    self.model.compartments, comp_vals, self.model._flows, flow_rates, computed_values, time
                )
                flow_rates[flow_idx] = net_flow


    def _get_rates(self, comp_vals: np.ndarray, time: float) -> Tuple[np.ndarray, np.ndarray]:
        """Calculates inter-compartmental flow rates for a given state and time, as well
        as the updated compartment values once these rate deltas have been applied

        Args:
            comp_vals (np.ndarray): The current state of the model compartments (ie. number of people)
            time (float): Time in model.times coordinates

        Returns:
            Tuple[np.ndarray, np.ndarray]: (comp_rates, flow_rates) where
                comp_rates is the rate of change of compartments, and
                flow_rates is the contribution of each flow to compartment rate of change
        """
        self._prepare_time_step(time, comp_vals)

        comp_rates = np.zeros(len(comp_vals))
        flow_rates = self._get_flow_rates(comp_vals, time)

        if self._pos_flow_map.size > 0:
            accumulate_positive_flow_contributions(flow_rates, comp_rates, self._pos_flow_map)

        if self._neg_flow_map.size > 0:
            accumulate_negative_flow_contributions(flow_rates, comp_rates, self._neg_flow_map)

        return comp_rates, flow_rates

    def get_compartment_rates(self, compartment_values: np.ndarray, time: float):
        """
        Interface for the ODE solver: this function is passed to solve_ode func and defines the dynamics of the model.
        Returns the rate of change of the compartment values for a given state and time.
        """
        comp_vals = self._clean_compartment_values(compartment_values)
        comp_rates, _ = self._get_rates(comp_vals, time)
        return comp_rates

    def get_flow_rates(self, compartment_values: np.ndarray, time: float):
        """
        Returns the contribution of each flow to compartment rate of change for a given state and time.
        """
        comp_vals = self._clean_compartment_values(compartment_values)
        _, flow_rates = self._get_rates(comp_vals, time)
        return flow_rates

"""
Additional functions - fast JIT specializations etc
"""

@numba.jit
def get_strain_infection_values(compartment_values, strain_compartment_infectiousness, compartment_category_map,
                                num_cats, mixing_matrix, category_populations):
    infected_values = compartment_values * strain_compartment_infectiousness
    infectious_populations = sparse_pairs_accum(
                compartment_category_map, infected_values, num_cats
            )
    infection_density = mixing_matrix @ infectious_populations
    category_prevalence = infectious_populations / category_populations
    infection_frequency = mixing_matrix @ category_prevalence

    return infection_density, infection_frequency




import random
import base64
from io import BytesIO
from typing import Callable, Optional, List

import igraph as ig
import networkx as nx
from networkx import drawing
import matplotlib.pyplot as plt
from PIL import Image

from .model import AgentModel

NodeFunc = Callable[[AgentModel, int], str]


def figure_to_buffer(fig) -> BytesIO:
    buf = BytesIO()
    fig.savefig(buf, format="png", bbox_inches="tight", pad_inches=0)
    return buf


def plots_to_gif_buf(bufs: List[BytesIO], duration: int) -> BytesIO:
    """
    Duration is time per frame, in ms.
    Use Gif Compressor if you want to share:

        https://gifcompressor.com/

    """
    images = [Image.open(buf) for buf in bufs]
    buf = BytesIO()
    images[0].save(
        buf,
        format="gif",
        save_all=True,
        append_images=images,
        optimize=False,
        duration=duration,
        loop=0,
    )
    buf.seek(0)
    return buf


def ipython_display_gif_buf(buf: BytesIO):
    """
    Displays a GIF buffer in Jupyter notebooks.
    """
    from IPython.display import display, HTML

    buf.seek(0)
    gif_base64_str = base64.b64encode(buf.read()).decode("utf-8")
    buf.seek(0)
    img_tag = f"<img src='data:image/gif;base64,{gif_base64_str}'>"
    display(HTML(img_tag))


def ipython_display_plot_buf(buf: BytesIO):
    """
    Displays a plots buffer in Jupyter notebooks.
    """
    from IPython.display import display

    buf.seek(0)
    img = Image.open(buf)
    display(img)


def ipython_display_plot_bufs(bufs: List[BytesIO]):
    """
    Displays an list of plots generated with `ipython_plot_model`
    display in Jupyter notebooks.
    """
    # Imported because not in release version
    import ipywidgets as wg

    def plot_timeslice(i):
        ipython_display_plot_buf(bufs[i])

    wg.interact(plot_timeslice, i=wg.IntSlider(min=0, max=len(bufs) - 1, step=1))


def plot_model_networks(
    model: AgentModel, seed: int = 1, get_node_color: Optional[NodeFunc] = None, draw_kwargs={}
):
    """
    Plots model networks onto a figure.
    Returns a matplotlib figure object.

    Mappings can be use to generate per-node kwargs.
    Kwargs can also be passed directly to draw_networkx.
    https://networkx.org/documentation/stable/reference/generated/networkx.drawing.nx_pylab.draw_networkx.html#networkx.drawing.nx_pylab.draw_networkx
    """
    # Combine all network graphs into one big graph.
    graphs = [model.networks.vals["graph"][i] for i in model.networks.query.ids()]
    graph = nx.compose_all(graphs)

    # Convert networkx graph into igraph format, use that to make a layout.
    random.seed(seed)
    igraph_graph = ig.Graph.from_networkx(graph)
    layout = igraph_graph.layout_fruchterman_reingold()

    # Plot actual graph
    plt.ioff()
    fig, ax = plt.subplots(1, 1, figsize=(12, 12), dpi=120)
    kwargs = {"ax": ax, "node_size": 50}

    if get_node_color:
        colors = []
        for node_id in graph.nodes.keys():
            rgb_color_tuple = get_node_color(model, node_id)
            rgb_color_tuple_normalised = tuple([n / 255.0 if n > 1 else n for n in rgb_color_tuple])
            colors.append(rgb_color_tuple_normalised)

        kwargs["node_color"] = colors

    with_labels = draw_kwargs.pop("with_labels", False)
    arrows = draw_kwargs.pop("arrows", True)
    kwargs = {**kwargs, **draw_kwargs}
    drawing.draw_networkx(graph, layout, arrows=arrows, with_labels=with_labels, **kwargs)
    return fig



from typing import Callable, Any, Optional
from abc import ABC, abstractmethod

import numpy as np
import networkx as nx


class BaseField(ABC):
    """
    A data field representing a component, which can be attached to an entity.
    Eg. the age (component/field) of a person (entity).
    """

    default: Optional[Any]
    distribution: Optional[Callable[[], Any]]

    @abstractmethod
    def __init__(self):
        self.default = None
        self.distribution = None

    @abstractmethod
    def setup(self, initial_number: int):
        """
        Create the initial data store for that field.
        """
        pass

    @abstractmethod
    def validate(self, value):
        """
        Asserts that the value is of the correct type for the field.
        """
        pass


class IntegerField(BaseField):
    def __init__(self, default: int = None, distribution: Callable[[], int] = None):
        has_dist = distribution is not None
        has_default = default is not None
        assert (
            has_dist or has_default
        ), "Must specify a default or a distribution for an IntegerField."
        assert not (
            has_dist and has_default
        ), "Cannot specify both a default and a distribution for an IntegerField."
        validate_value = distribution() if has_dist else default
        self.validate(validate_value)
        self.default = default
        self.distribution = distribution

    def validate(self, value):
        try:
            assert type(value) in (int, np.int32, np.int64), "IntegerField values must be of int type."
        except:
            print(type(value),value)
            raise

    def setup(self, initial_number: int):
        if self.default:
            arr = np.ones(initial_number, dtype=np.int)
            arr *= self.default
        else:
            arr = np.zeros(initial_number, dtype=np.int)
            for i in range(initial_number):
                arr[i] = self.distribution()

        return arr


class GraphField(BaseField):
    """
    Represents an undirected graph using a NetworkX Graph object.
    """

    def __init__(self):
        super().__init__()
        self.distribution = self._get_default

    def _get_default(self):
        return nx.Graph()

    def setup(self, initial_number: int):
        arr = np.empty(initial_number, dtype=np.object)
        for i in range(initial_number):
            arr[i] = self._get_default()

        return arr

    def validate(self, value):
        assert type(value) is nx.Graph, "GraphField values must be a Networkx Graph."